const findIndex = (array: any[], value: any, comparator: Function) => {
  let start = 0
  let end = array.length
  while (start < end) {
    const middle = (start + end) >>> 1
    if (comparator(array[middle], value) <= 0) {
      start = middle + 1
    } else {
      end = middle
    }
  }
  return start
}

class PriorityQueue {
  options: {
    initialValues: any[]
  }
  comparator: (a: any, b: any) => number
  data: any[]
  constructor(options: {
    comparator: (a: any, b: any) => number
    initialValues: any[]
  }) {
    this.options = options
    this.comparator = options.comparator
    this.data = (options.initialValues && options.initialValues.slice()) || []
    this.data.sort(this.comparator).reverse()
  }

  queue(value: any) {
    const index = findIndex(this.data, value, this.comparator)
    this.data.splice(index, 0, value)
  }

  dequeue() {
    return this.data.pop()
  }

  peek() {
    return this.data[this.data.length - 1]
  }

  find(value: any) {
    const index = findIndex(this.data, value, this.comparator) - 1
    return index >= 0 && this.comparator(this.data[index], value) === 0 ? index : -1
  }

  remove(value: any) {
    const index = this.find(value)
    if (index >= 0) {
      this.data.splice(index, 1)
      return true
    }
    return false
  }

  removeAllMatching(test: Function, callback: Function) {
    let count = 0
    for (let i = this.data.length - 1; i >= 0; i--) {
      if (test(this.data[i])) {
        const value = this.data.splice(i, 1)[0]
        if (callback) {
          callback(value)
        }
        count++
      }
    }
    return count
  }
}
